@@ -20,7 +20,7 @@ allow running on Python 2.
 """
 
 import re
-from typing import Any, NamedTuple, Optional, Union
+from typing import Any, Dict, NamedTuple, Optional, Union
 
 # _Instruction.tos_str.__doc__ = (
 #     "If not None, a string representation of the top of the stack (TOS)"
@@ -126,11 +126,8 @@ class Instruction(NamedTuple):
       arg:     Optional numeric argument to operation (if any). Otherwise, None.
 
       argval:  resolved arg value (if known). Otherwise, the same as ``arg``.
-      argrepr: human-readable description of operation argument.
 
-      tos_str:      If not None, a string representation of the top of the stack (TOS).
-                    This is obtained by scanning previous instructions and
-                    using information there and in their ``tos_str`` fields.
+      argrepr: human-readable description of operation argument.
 
       positions: Optional dis.Positions object holding the start and end locations that
                  are covered by this instruction. This not implemented yet.
@@ -148,6 +145,14 @@ class Instruction(NamedTuple):
                     instruction. Note conditionals are in this category, but
                     returns, raise, and unconditional jumps are not.
 
+      Note: the following fields have to appear in the order below and be at the end.
+      disassembly may replace (delete and insert) an instruction, and it assumes
+      the ending fields are as follows:
+
+      tos_str:      If not None, a string representation of the top of the stack (TOS).
+                    This is obtained by scanning previous instructions and
+                    using information there and in their ``tos_str`` fields.
+
       start_offset: if not None the instruction with the lowest offset that
                     pushes a stack entry that is consume by this opcode
     """
@@ -170,7 +175,7 @@ class Instruction(NamedTuple):
     argval: Any
 
     # String representation of argval if argval is not None.
-    argrepr: str
+    argrepr: Optional[str]
 
     # Offset of the instruction
     offset: int
@@ -226,18 +231,25 @@ class Instruction(NamedTuple):
     def disassemble(
         self,
         opc,
+        line_starts: Optional[Dict[int, int]] = None,
         lineno_width=3,
         mark_as_current=False,
         asm_format="classic",
         instructions=[],
     ):
-        """Format instruction details for inclusion in disassembly output
+        """
+        Format instruction details for inclusion in disassembly output.
 
-        *lineno_width* sets the width of the line number field (0 omits it)
-        *mark_as_current* inserts a '-->' marker arrow as part of the line
+        ``line_starts`` when it exists is a dictionary mapping a bytecode offsets to
+        line numbers.
+
+        ``lineno_width`` sets the width of the line number field (0 omits it)
+
+        ``mark_as_current`` inserts a '-->' marker arrow as part of the line.
         """
         fields = []
         indexed_operand = frozenset(["name", "local", "compare", "free"])
+        opcode = self.opcode
 
         # Column: Source code line number
         if lineno_width:
@@ -277,7 +289,7 @@ class Instruction(NamedTuple):
 
         # Column: Instruction bytes
         if asm_format in ("extended-bytes", "bytes"):
-            hex_bytecode = "|%02x" % self.opcode
+            hex_bytecode = "|%02x" % opcode
             if self.inst_size == 1:
                 # Not 3.6 or later
                 hex_bytecode += " " * (2 * 3)
@@ -287,7 +299,7 @@ class Instruction(NamedTuple):
                     hex_bytecode += " %02x" % (self.arg % 256)
                 else:
                     hex_bytecode += " 00"
-            elif self.inst_size == 3:
+            elif self.inst_size == 3 and self.arg is not None:
                 # Not 3.6 or later
                 hex_bytecode += " %02x %02x" % divmod(self.arg, 256)
 
@@ -298,13 +310,14 @@ class Instruction(NamedTuple):
 
         # Column: Opcode argument
         if self.arg is not None:
+
             argrepr = self.argrepr
             # The ``argrepr`` value when the instruction was created
             # generally has all the information we require.  However,
             # for "asm" format, want additional explicit information
             # linking operands to tables.
             if asm_format == "asm":
-                if self.optype in ("jabs", "jrel"):
+                if self.is_jump() and self.argrepr is not None:
                     assert self.argrepr.startswith("to ")
                     jump_target = self.argrepr[len("to ") :]
                     fields.append("L" + jump_target)
@@ -312,23 +325,36 @@ class Instruction(NamedTuple):
                     fields.append(repr(self.arg))
                     fields.append("(%s)" % argrepr)
                     argrepr = None
-                elif self.optype == "const" and not re.search(r"\s", argrepr):
-                    fields.append(repr(self.arg))
-                    fields.append("(%s)" % argrepr)
-                    argrepr = None
-                elif self.optype == "const" and not re.search(r"\s", argrepr):
+                elif (
+                    self.optype == "const"
+                    and argrepr is not None
+                    and not re.search(r"\s", argrepr)
+                ):
                     fields.append(repr(self.arg))
                     fields.append("(%s)" % argrepr)
                     argrepr = None
                 else:
                     fields.append(repr(self.arg))
             elif asm_format in ("extended", "extended-bytes"):
-                op = self.opcode
                 if (
-                    hasattr(opc, "opcode_extended_fmt")
-                    and opc.opname[op] in opc.opcode_extended_fmt
+                    self.is_jump()
+                    and line_starts is not None
+                    and line_starts.get(self.argval) is not None
                 ):
-                    new_repr = opc.opcode_extended_fmt[opc.opname[op]](
+                    new_instruction = list(self)
+                    new_instruction[-2] = f"To line {line_starts[self.argval]}"
+                    # Here and below we use self.__class__ instead of Instruction
+                    # so that other kinds of compatible namedtuple Instructions
+                    # can be used. In particular, the control-flow project
+                    # defines such an ExtendedInstruction namedtuple
+                    self = self.__class__(*new_instruction)
+                    del instructions[-1]
+                    instructions.append(self)
+                elif (
+                    hasattr(opc, "opcode_extended_fmt")
+                    and self.opname in opc.opcode_extended_fmt
+                ):
+                    new_repr = opc.opcode_extended_fmt.get(self.opname, lambda opc, instr: None)(
                         opc, list(reversed(instructions))
                     )
                     start_offset = None
@@ -337,13 +363,22 @@ class Instruction(NamedTuple):
                     if new_repr:
                         # Add tos_str info to tos_str field of instruction.
                         # This the last field in instruction.
-                        new_instruction = list(instructions[-1])
-                        new_instruction[-2] = new_repr
+                        new_instruction = list(self)
                         new_instruction[-1] = start_offset
+                        new_instruction[-2] = new_repr
                         del instructions[-1]
-                        instructions.append(Instruction(*new_instruction))
+                        # See comment above abut the use of self.__class__
+                        self = self.__class__(*new_instruction)
+                        instructions.append(self)
                         argrepr = new_repr
-                        start_offset = start_offset
+                elif opcode in opc.nullaryloadop:
+                    new_instruction = list(self)
+                    new_instruction[-2] = self.argrepr
+                    start_offset = new_instruction[-1] = self.offset
+                    del instructions[-1]
+                    # See comment above abut the use of self.__class__
+                    self = self.__class__(*new_instruction)
+                    instructions.append(self)
                 pass
             if not argrepr:
                 if asm_format != "asm" or self.opname == "MAKE_FUNCTION":
@@ -352,40 +387,51 @@ class Instruction(NamedTuple):
             else:
                 # Column: Opcode argument details
                 if len(instructions) > 0:
-                    argval = instructions[-1].argval
-                    if instructions[-1].tos_str is None or (
-                        self.argrepr is not None
-                        and self.argrepr == instructions[-1].tos_str
-                    ):
+                    argval = self.argval
+                    if self.tos_str is None:
                         fields.append(f"({self.argrepr})")
                     else:
-                        prefix = "" if self.argrepr is None else f"({self.argrepr}) ; "
-                        if self.opcode in opc.operator_set:
+                        if self.optype in ("vargs", "encoded_arg"):
+                            prefix = f"{self.argval} ; "
+                        elif self.argrepr is None:
+                            prefix = ""
+                        else:
+                            prefix = f"({self.argrepr}) ; "
+                        if opcode in opc.operator_set | opc.callop:
                             prefix += "TOS = "
-                        fields.append(f"{prefix}{instructions[-1].tos_str}")
+                        fields.append(f"{prefix}{self.tos_str}")
                     pass
+                elif self.argrepr is not None:
+                    fields.append(self.argrepr)
                 pass
             pass
         elif asm_format in ("extended", "extended-bytes"):
-            op = self.opcode
             if (
                 hasattr(opc, "opcode_extended_fmt")
-                and opc.opname[op] in opc.opcode_extended_fmt
+                and self.opname in opc.opcode_extended_fmt
             ):
-                new_repr, start_offset = opc.opcode_extended_fmt[opc.opname[op]](
+                new_repr, start_offset = opc.opcode_extended_fmt.get(self.opname, (None, 0))(
                     opc, list(reversed(instructions))
                 )
                 if new_repr:
-                    new_instruction = list(instructions[-1])
+                    new_instruction = list(self)
                     new_instruction[-2] = new_repr
                     new_instruction[-1] = start_offset
                     del instructions[-1]
-                    instructions.append(Instruction(*new_instruction))
-                    argval = instructions[-1].argval
+                    # See comment above abut the use of self.__class__
+                    instructions.append(self.__class__(*new_instruction))
+                    argval = self.argval
                     prefix = "" if argval is None else f"({argval}) | "
                     if self.opcode in opc.operator_set:
                         prefix += "TOS = "
                     fields.append(f"{prefix}{new_repr}")
+                pass
+            elif (
+                hasattr(opc, "opcode_arg_fmt")
+                and self.opname in opc.opcode_arg_fmt
+            ) and self.argrepr is not None:
+                fields.append(self.argrepr)
+                pass
             pass
 
         return " ".join(fields).rstrip()
